///////////////////////////////////////////////////////////////////////////////
//
// Jail
// A chrooted login shell implementation
// Released under GNU Public License v.2.0
// See LICENSE file for details
//
// Juan M. Casillas <assman@gsyc.inf.uc3m.es>
// 3/April/ 2001 09:22
//
///////////////////////////////////////////////////////////////////////////////
//
// $Id: jail.c,v 1.7 2001/04/16 08:47:04 assman Exp $
//
// $Log: jail.c,v $
// Revision 1.7  2001/04/16 08:47:04  assman
// fixed a memory leak bug, fixed the directory read bug also
//
// Revision 1.5  2001/04/10 15:42:13  assman
// added documentation, change the Makefile to check de dependencies
// with mkenv.sh added TODO file, removed debug messages in the code
//
// Revision 1.4  2001/04/10 10:33:23  assman
// added CHANGELOG, INSTALL files, moved mkenv.sh for each platform,
// added new info to the documentation
//
// Revision 1.2  2001/04/03 10:28:36  assman
// removed binary file added by error
// added CVS tags
//
// 11/April/2001 --  I lost the cvs in my portable PC :(
//                   We are going to do a major cleanup in the code
//                   and improve some parts of Jail
//
// 13/April/2001 --  The user can choose the shell he wants using the
//                   "chrooted" /etc/passwd file. Major improvements in
//                   mkenv.sh, now it supports guessing the libraries
//                   it will use (we need to fine-tune this process using
//                   strace, this program is not installed on my
//                   portable PC :( )
//
///////////////////////////////////////////////////////////////////////////////
//
// NOTES
//
// File must be owned by root and have the SUID permissions
// (4755 runs fine) to allow proper operation. Also the chroot
// environment must be setup right. Se documentation files (README, INSTALL)
// for configuration and instalation
//
///////////////////////////////////////////////////////////////////////////////
//
// BUGS
//
// some curious happen when trying to run jail in an IRIX 6.5
// when it starts, it becomes killed by itself. Tracing it, I
// think is a problem with the execve() system call. 
// Work arrounds for this platform are wellcome.
//
///////////////////////////////////////////////////////////////////////////////

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <sys/types.h>
#include <pwd.h>

//////////////////////////////////////////////////////////////////////////////
//
// DEFINE section
//
// VERSION           The CVS version of the program. This is not the 
//                   release tag, is the revision tag. You shouldn't change
//                   it
//
// HOME_VAR          When building user's homedir path, we need to prefix
//                   the HOME variable with this header.
//                 
// SHELL_VAR         When building user's shell variable, we need to prefix
//                   the SHELL variable with this header.
//
// BANNER            Some info about Jail program
//
//////////////////////////////////////////////////////////////////////////////

#define SHELL "/bin/bash"

#define VERSION "$Id: jail.c,v 1.7 2001/04/16 08:47:04 assman Exp $"

#define BANNER "\nJail version " VERSION "\n\n" \
	       "This program is released under the GNU License\n" \
	       "Juan M. Casillas <assman@gsyc.inf.uc3m.es>\n" \
	       "See http://www.gsyc.inf.uc3m.es/~assman/jail\n" \
	       "for more info\n\n"

#define HOME_VAR "HOME="
#define SHELL_VAR "SHELL="

//////////////////////////////////////////////////////////////////////////////
//
// init_env()
// Duplicate a string and save it into the pointer pointed by ptr
//
// INPUT args:
//
//       **ptr        the pointer that will point to the dup string
//       *str         a pointer to the string
//
// OUTPUT args:
//
//       noting
//
//////////////////////////////////////////////////////////////////////////////

void init_env(char **ptr, char *str) {

    *ptr = (char *)malloc(strlen(str)+1);
    memset(*ptr,0,strlen(str)+1);
    strcpy(*ptr,str);
}

//////////////////////////////////////////////////////////////////////////////
//
// process_args()
// Process the command line arguments
//
// Only accepts one argument, --version, else exits with an error code
//
// INPUT args:
//
//       argc         an integer that counts the number of arguments
//       *argv[]      an array of pointers pointing to the arguments
//
// OUTPUT args:
//
//       1 if none argument found
//       else, exits with a 0 value if any argument has been succesfully
//       parsed, exits with 1 if any parameter other than --version is
//       passed
//
//////////////////////////////////////////////////////////////////////////////

int process_args(int argc, char *argv[]) {
    if (argc > 1 && !strcmp(argv[1],"--version")) {
	fprintf(stderr,BANNER 
		"params:\n\n"
		"\t--version: show this version\n\n"
		"this program is mainly used for login users into the system\n"
		"so interactive use is not useful, as you can see :)\n\n"
		"Info ends, so bailing out.\n\n");
	exit(0);
    }
    if (argc > 1) {
	fprintf(stderr,BANNER
		"error: You can't pass arguments to jail\n"
		"bailing out.\n\n");
	exit(1);
    }
    return(0);
}

//////////////////////////////////////////////////////////////////////////////
//
// main()
// main execution loop
// 
// INPUT args:
//
//       argc           an integer that counts the number of arguments
//       *argv[]        an array of pointers pointing to the arguments
//       *env[]         an array of pointers pointing to the enviroment 
//                      variables passed from the caller program (usually, 
//                      a login daemon, just like telnetd, sshd, etc).
//
// OUTPUT args:
// 
//       Jail nevers returns (is overwritten by another process)
//       if some error ocurrs then the following values are returned:
//
//             1        A non valid argument has been passed in the command
//                      line to jail. Usually, an argument different from
//                      --version
//             
//             2        Errno in the following cases:
// 
//                      1) can't do a chdir()
//                      2) can't do a chroot()
//                      3) can't do a execve()
//             
//            -1        If we reached the last line of jail, that was a 
//                      really big problem, because jail should never 
//                      returns
//
//////////////////////////////////////////////////////////////////////////////

int main(int argc, char *argv[], char *env[]) {
    int ret;
    char **lenv;
    char **larg;
    int n;
    struct passwd *pwdent,*cpwdent;
    int uid, gid;
    char *home_dir;
    char *shell_path;
    char *ptr_home_dir;
    int argc_cnt;

    process_args(argc,argv);

    //
    // Get user id and group id
    // become root
    // get passwd file info for the user
    //

    uid = getuid();
    gid = getgid();

    if (setgid(0) || setuid(0) || getuid() != geteuid() || 
	getgid() != getegid()) {
	fprintf(stderr,"I can't do the chroot() because Im not root");
    }

    pwdent = getpwuid(uid);	

    //
    // count elements and generate space for the array
    // then, save all the environment variables but do
    // nothing with HOME and SHELL
    //

    n=0;
    while (env[n]) n++;  
    lenv = malloc(sizeof(char *)*n);
    
    n=argc_cnt=0;
    while (env[n]) {
	if ((strstr(env[n],"HOME")==NULL) && 
	    (strstr(env[n],"SHELL")==NULL)) {
	    init_env(&lenv[argc_cnt],env[n]); 
	    //printf("%s\n",lenv[j]);
	    argc_cnt++;
	}
	n++;
    }	
    
    //
    // chdir & Chroot the users' directory readed before in pwent
    // Read again the passwd file info, that will be /virtual/etc/passwd
    // if the two entries are the same, do nothing, else, chdir to the
    // new directory, and save this info in the new environment argument
    // array.
    // 
    // next, get the shell entry from the chrooted /etc/passwd file
    // and set it as shell variable
    //

    ret = chdir(pwdent->pw_dir);	
    if (ret) {
	perror("chdir() ");
	return(ret);
    }
	
    ret = chroot(pwdent->pw_dir);
    if (ret) { 
	perror("chroot() ");
	return(ret);
    }

    cpwdent = getpwuid(uid);

    if (strcmp(pwdent->pw_dir,cpwdent->pw_dir)) 
	ptr_home_dir = cpwdent->pw_dir;
    else
	ptr_home_dir = pwdent->pw_dir;

    n = strlen(HOME_VAR)+strlen(ptr_home_dir)+1;
    home_dir = (char *)malloc(n);
    memset(home_dir,0,n);
    strcpy(home_dir,HOME_VAR);
    strcat(home_dir,ptr_home_dir);

    n = strlen(SHELL_VAR)+strlen(cpwdent->pw_shell)+1;
    shell_path = (char *)malloc(n);
    memset(shell_path,0,n);
    strcpy(shell_path,SHELL_VAR);
    strcat(shell_path,cpwdent->pw_shell);

    init_env(&lenv[argc_cnt],home_dir); argc_cnt++;
    init_env(&lenv[argc_cnt],shell_path); argc_cnt++;	
    lenv[argc_cnt]=NULL;

    larg = malloc(sizeof(char *)*4); 
    init_env(&larg[0],argv[0]);
    init_env(&larg[1],"-");
    larg[2]=NULL;

    //
    // Finally become the user, 
    // chdir to the user home directory,
    // clean all the structures and
    // call the shell overriding current process space
    //

    setgid(gid);
    setuid(uid);

    ret = chdir(ptr_home_dir);	
    if (ret) {
	perror("[chrooted]chdir() ");
	return(ret);
    }

    free(home_dir);
    free(shell_path);

    ret = execve(SHELL,larg,lenv);
    if (ret) { 
	perror("execve() ");
	return(errno);
    }

    //
    // not reached and if so, we are in trouble :(
    //

    return(-1); 
}

    
