#!/bin/bash

# -----------------------------------------------------------------------------
#
# mkenv.sh
# generate chroot & user environment
# A member of Jail Chroot Project.
# 
# Juan M. Casillas
# assman@gsyc.inf.uc3m.es
#
# script that generates the chroot environment based on the platform
# and the home directory of the user. This release supports various
# platforms, and it is extensible and highly configurable.
#
# Complete rewrite of the script
# mkenv.sh now customizes user directory also (automatically edits 
# the /etc/passwd chrooted file)
#
# $Id: mkenv.sh,v 1.4 2001/05/03 07:47:03 assman Exp $
#
# $Log: mkenv.sh,v $
# Revision 1.4  2001/05/03 07:47:03  assman
# modified mkenv now, it guess where the files are
#
# Revision 1.3  2001/04/16 08:46:05  assman
# Removed Path Splitting, changed it for a new feature based on
# read the chrooted /etc/passwd. Added new features in mkenv.sh,
# now it can guess about the library and file dependencies of the
# binaries. Removed platform-specific mkenv.sh files
#
#
# 17/April/2001
# I have touched the code, now mkenv is platform independent.
# no platform is required
#
# -----------------------------------------------------------------------------

FILES=""
DEVICES=""
LIB=""

# -----------------------------------------------------------------------------
#
# lib_dep()
# generate library dependences of a program, so 
# we can adapt dinamically to any platform. 
#
# depends on ldd program and its output
# using ldd of a debian potato 2.2
# depends on strace
#
# variable LIBS, FILES and DEVICES are modified
# inside this function
#
# -----------------------------------------------------------------------------

function lib_dep() {
LDD="ldd -v"

l_LIBS="";
l_DEVICES="";
l_FILES="";

l_LIBS=`$LDD $1 | grep =\> | awk -F"=>" '{ print $2}' | awk '{print $1}'`

deps=`(strace $1 </dev/null 2>&1) | grep open | \
       grep -v "No such file or directory" | \
       awk -F\( '{print $2}' | awk -F, '{print $1}' | tr -d '"'`

for i in $deps; do

 if [ -b $i ]; then
    major=`\ls -la $i | awk '{ print $5}' | tr -d ','`
    minor=`\ls -la $i | awk '{ print $6}'`
    #echo "$i is a block special device: $major $minor"
    l_DEVICES=`echo "$l_DEVICES $i:b:$major:$minor"`
 else 
   if [ -c $i ]; then
    major=`\ls -la $i | awk '{ print $5}' | tr -d ','`
    minor=`\ls -la $i | awk '{ print $6}'`
    #echo "$i is a char special device: $major $minor"
    l_DEVICES=`echo "$l_DEVICES $i:c:$major:$minor"`
   else
     if [ -f $i ]; then
      #echo "$i is a normal file: `ls -la $i`" 
      #
      # now, if it has a lib in the name, put it into the variable
      # libs; else, put it into "files" section. Be careful about
      # full directories
      # XXX
      res=`echo $i | grep lib`
      if [ "$res" != "" ]; then
         l_LIBS=`echo "$l_LIBS $i"`;
      else
         l_FILES=`echo "$l_FILES $i"`;
      fi
      

     fi
   fi
 fi

done



LIBS=`echo "$LIBS $l_LIBS"`
DEVICES=`echo "$DEVICES $l_DEVICES"`
FILES=`echo "$FILES $l_FILES"`

}

# -----------------------------------------------------------------------------
#
# more things here.
# show basic usage
#
# -----------------------------------------------------------------------------


# -----------------------------------------------------------------------------
#
# usage()
# show basic usage
#
# -----------------------------------------------------------------------------

function usage() {
echo "
mkenv.sh
A component of Jail
http://www.gsyc.inf.uc3m.es/~jail/
Juan M. Casillas <assman@gsyc.inf.uc3m.es>

usage: $0 chrootdir userdir usershell username when:
	  
  chrootdir       is the directory to chroot
  userdir         is the userdir inside the chroot environment
  usershell       the shell that the user will use
  username        the user id
  "
}

# -----------------------------------------------------------------------------
#
# copy_create()
# copy a file prepending to the path $2 and build the directory if 
# it doesn't exist It does some checks also to test the validity of
# the path
#
# Don't overwrite custom files: /etc/passwd /etc/shadow /etc/group
# under the chrooted directory
#
# -----------------------------------------------------------------------------

function copy_create() {

base=`dirname $1`

dst_dir="$2/$base"

if [ ! -d $dst_dir ]; then

  mkdir -p $dst_dir;

fi

case $1 in 
 /etc/passwd | /etc/shadow | /etc/group) 
	if [ ! -e $2/$1 ]; then
	   cp -r $1 $dst_dir
	else
	   echo "WARNING: Can't overwrite $1" ;
	fi
	;;
	*) cp -r $1 $dst_dir
esac

}

# -----------------------------------------------------------------------------
#
# gen_user()
#
# generate a user directory inside the chroot environment
# 1) create the directory if applicable
# 2) change the passwd file of the passwd file inside the chrooted environment
#
# $1 chrooted environment
# $2 user home directory
# $3 user shell
# $4 user name
#
# -----------------------------------------------------------------------------

function gen_user() {

entry=`cat $1/etc/passwd | grep -e "^$4:"`
tmpasswd=`mktemp /tmp/mkenv.XXXXXX`
cat $1/etc/passwd | grep -v -e "^$4:" > $tmpasswd

#
# if userdir is equal to '-', don't change it
#

if [ "$2" != "-" ]; then 
 val=`echo "$2\n" | sed -e "s/\//\\\\\\\\\//g" | sed -e "s/\\\\\\n//g"`
 entry=`echo $entry | awk -F: '{printf ("%s:%s:%s:%s:%s:__HOME__:%s\n",  \
				$1, $2, $3, $4, $5, $7); }' | \
	sed -e "s/__HOME__/$val/g"`

    #
    # create the directory entry, chown to the user / group values, and bailout
    #

    u_uid=`echo $entry | awk -F: '{ print $3}'`
    u_gid=`echo $entry | awk -F: '{ print $4}'`

    mkdir -p $1/$2
    chown -R $u_uid:$u_gid $1/$2

else
 entry=`echo $entry | awk -F: '{printf ("%s:%s:%s:%s:%s:/:%s\n",  \
				$1, $2, $3, $4, $5, $7); }'`
fi

#
# change the shell
#

 val=`echo "$3\n" | sed -e "s/\//\\\\\\\\\//g" | sed -e "s/\\\\\\n//g"`
 entry=`echo $entry | awk -F: '{printf ("%s:%s:%s:%s:%s:%s:__SHELL__\n",  \
				$1, $2, $3, $4, $5, $6); }' | \
	sed -e "s/__SHELL__/$val/g"`

#
# build the passwd file and save it to its location
# clear temporal files and go out
#

(cat $tmpasswd; echo $entry) > $1/etc/passwd

rm $tmpasswd
}


# =============================================================================
# 
# for each platform, generate five functions
#
# 1) preinstall step ---------------> pre_install_<platform>
# 2) copy program files ------------> program_<platform>
# 3) copy files, extra libs, etc ---> files_<platform>
# 4) make devices ------------------> devices_<platform>
# 5) generate user directory
# 6) postinstall step --------------> post_install_<platform>
#
# Superseeded all the code. Only allow unix
# for future compatibility
#
# =============================================================================

# 
# debian section
# 

function pre_install_unix() {

mkdir -p $1/dev $1/tmp $1/var/tmp

DIRS="/dev /tmp /var/tmp";
echo $DIRS;
}    

function programs_unix() {

PROGS="bash sh cat grep id ls \
       mkdir mv cp ln mv pwd rm \
       rmdir tail more touch vi \
       head"

echo $PROGS
}


function files_unix() {

#l_FILES="/etc/terminfo /etc/group /etc/localtime /etc/nsswitch.conf \
#         /etc/passwd /etc/shadow  /lib/libnss_compat.so.2 /lib/libnsl.so.1"
l_FILES=""

FILES=`echo "$FILES $l_FILES"`
}


function devices_unix() {

#l_DEVICES="/dev/zero:c:1:5 /dev/null:c:1:3"
l_DEVICES=""

DEVICES=`echo "$DEVICES $l_DEVICES"`
#echo $DEVICES;    
}

function post_install_unix() {

chmod -R 4777 $1/tmp $1/var/tmp
rm -rf $1/tmp/* $1/var/tmp/*

DIRS="/dev /tmp /var/tmp";
echo $DIRS;
}    


# -----------------------------------------------------------------------------
#
# main loop
# 
# check for parameters
# choose architecture 
# Call platform funcions
# issue the copy
# platform is superseeded
#
# -----------------------------------------------------------------------------

if [ $# -ne 4 ]; then
	usage;
	exit 1;
fi

HDIR=$1
UDIR=$2
USHL=$3
UNME=$4

#
# Platform is superseeded
#

PLAT="unix"

#
# test platform for correctness
#

case $PLAT in
 debian_2_2) ;;
 redhat_6_2) ;;
caldera_7_0) ;;
       unix) ;;
          *) usage
             exit 1
esac

echo "Making $HDIR chrooted environment for platform $PLAT"

# 
# call functions
#

echo "Doing pre_install_$PLAT()"

res=`pre_install_$PLAT $HDIR`

#
# get all the programs and build library dependencies
# into a variable, then copy the files and the libraries
# to the directories.
#

echo "Doing programs_$PLAT()"

for i in `programs_$PLAT`; do

  #
  # we modify inside lib_dep the following global variables:
  # LIBS FILES DEVICES

  echo -en "\tguessing $i ..."
  tp=`which $i`

  if [ "$tp" != "" ]; then
    lib_dep `which $tp`
    echo $tp
   else
    echo "Failed!"
   fi


done

LIBS=`echo $LIBS | tr ' ' '\n' | sort | uniq | grep -v ^$ | tr -d ' '`
FILES=`echo $FILES | tr ' ' '\n' | sort | uniq | grep -v ^$ | tr -d ' '`
DEVICES=`echo $DEVICES | tr ' ' '\n' | sort | uniq | grep -v ^$ | tr -d ' '`

#
# copy libs & programs
#

echo -en "\tInstalling Libraries ";

for i in $LIBS; do
 copy_create $i $HDIR
 echo -en "."
done

echo " Done."

echo -en "\tInstalling programs ".

for i in `programs_$PLAT`; do
tp=`which $i`

  if [ "$tp" != "" ]; then 
    copy_create $tp $HDIR
    echo -en "."
  else
    echo -en "x"
  fi

done

echo " Done."

#
# calling files (copy that files that we can't guess about them
# in the future release, we can guess them using strace
#

echo "Doing files_$PLAT()"
echo -en "\tInstalling files ";

#
# call files_$PLAT so we add to the variable $FILES all the files
# that need to be copied
#

files_$PLAT 

for i in $FILES; do
 copy_create $i $HDIR
 echo -en "."
done

echo " Done."

#
# making devices
#

echo "Doing devices_$PLAT()"
devices_$PLAT

for i in $DEVICES; do
 d_name=`echo $i | awk -F: '{print $1}'`
 d_type=`echo $i | awk -F: '{print $2}'`
 d_major=`echo $i | awk -F: '{print $3}'`
 d_minor=`echo $i | awk -F: '{print $4}'`

 mknod ${HDIR}${d_name} $d_type $d_major $d_minor

done

#
# generating the userdir 
#

echo "Generating user directory"
gen_user $HDIR $UDIR $USHL $UNME

#
# calling post-install
#

echo "Doing post_install_$PLAT()"
res=`post_install_$PLAT $HDIR`

#
# end the script
#

echo "Chrooted environment for platform $PLAT in $HDIR Done"

exit 0;

